"""
/***************************************************************************
 QMarxanDialog
                                 A QGIS plugin
 Create grid, measure input layers, create Marxan input layers, import results
                             -------------------
        begin                : 2011-10-11
        copyright            : (C) 2011 by Apropos Information Systems Inc.
        email                : tsw@aproposinfosystems.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from PyQt4 import QtCore, QtGui
from qgis.core import *
import qgis.utils
from ui_qm_calc import Ui_qm_calc
import os, time

# create the dialog for zoom to point
class qmCalc(QtGui.QDialog):

    def __init__(self, iface):
        self.iface = iface

        # Import required plugins and if missing show the plugin's name
        import sys
        ftPath = QtCore.QFileInfo(QgsApplication.qgisUserDbFilePath()).path() + '/python/plugins/fTools'
        if not ftPath in sys.path:
            sys.path.append(ftPath)
        req_plugin = "ftools 0.5.10"
        try:
            import ftools_utils
        except:
            QtGui.QMessageBox.information(self, self.tr("Missing Plugins"), 
                    self.tr("Missing plugin: %s" % req_plugin))

        QtGui.QDialog.__init__(self)
        # Set up the user interface from Designer.
        self.ui = Ui_qm_calc()
        self.ui.setupUi(self)
        self.setFixedSize(510,640)
        #
        self.setWindowTitle(self.tr("Calculate Planning Grid Conservation Factor Values"))
        # add input layers to select planning grid
        layers = ftools_utils.getLayerNames([QGis.Polygon])
        self.ui.cbxPlanningGrid.addItems(layers)
        self.setMeasurePoints()
        self.setMeasureFields(self.ui.cbxPlanningGrid.currentText())
        self.setOutputField()
        self.ui.buttonOk = self.ui.buttonBox.button( QtGui.QDialogButtonBox.Ok )
        self.ui.rdoSum.setChecked(True)
        self.ui.rdoMeasure.setChecked(True)
        
        # connect methods to button interactions
        # selecting type of layer to measure
        QtCore.QObject.connect(self.ui.rdoPoints, QtCore.SIGNAL("clicked()"), self.setMeasurePoints)
        QtCore.QObject.connect(self.ui.rdoLines, QtCore.SIGNAL("clicked()"), self.setMeasureLines)
        QtCore.QObject.connect(self.ui.rdoAreas, QtCore.SIGNAL("clicked()"), self.setMeasurePolygons)
        # selecting measure layer
        QtCore.QObject.connect(self.ui.cbxPlanningGrid, QtCore.SIGNAL("currentIndexChanged(QString)"), self.setMeasureFields)
        # select single or multiple variale output
        QtCore.QObject.connect(self.ui.rdoCalcSingle, QtCore.SIGNAL("clicked()"), self.setSingle)
        QtCore.QObject.connect(self.ui.rdoCalcMulti, QtCore.SIGNAL("clicked()"), self.setMulti)
        # selecting measure or calculation
        QtCore.QObject.connect(self.ui.rdoMeasure, QtCore.SIGNAL("clicked()"), self.setMeasure)
        QtCore.QObject.connect(self.ui.rdoCalculate, QtCore.SIGNAL("clicked()"), self.setCalculate)
        QtCore.QObject.connect(self.ui.rdoPresence, QtCore.SIGNAL("clicked()"), self.setPresence)
        QtCore.QObject.connect(self.ui.cbxMeasureLayer, QtCore.SIGNAL("currentIndexChanged(QString)"), self.setCalcFields)
        # selecting field name
        QtCore.QObject.connect(self.ui.cbxSelectField, QtCore.SIGNAL("currentIndexChanged(QString)"), self.setOutputField)

    def setMeasure(self):
        self.ui.lblCalcField.setDisabled(True)
        self.ui.cbxCalcField.setDisabled(True)
        if not self.ui.lblMultipleAction.isEnabled():
            self.ui.lblMultipleAction.setEnabled(True)
            self.ui.rdoSum.setEnabled(True)
            self.ui.rdoMean.setEnabled(True)
            self.ui.rdoMax.setEnabled(True)
            self.ui.rdoMin.setEnabled(True)
            self.ui.rdoCount.setEnabled(True)

    def setSingle(self):
        # switch buttons
        self.ui.rdoCalcSingle.setChecked(True)
        self.ui.rdoCalcMulti.setChecked(False)
        # enable insection action options
        self.ui.lblMeasureOrCalc.setEnabled(True)
        self.ui.rdoMeasure.setEnabled(True)
        self.ui.rdoPresence.setEnabled(True)
        self.ui.rdoCalculate.setEnabled(True)
        self.setMeasure()
        self.ui.lblCalcField.setText('Select Field for Calculation')
        self.ui.lblSelectField.setEnabled(True)
        self.ui.cbxSelectField.setEnabled(True)
        self.ui.lblNewField.setText('Enter New Measure Field Name')
        self.ui.leNewField.setMaxLength(10)

    def setMulti(self):
        self.ui.rdoCalcSingle.setChecked(False)
        self.ui.rdoCalcMulti.setChecked(True)
        # disable insection action options
        self.ui.lblMeasureOrCalc.setEnabled(False)
        self.ui.rdoMeasure.setChecked(True)
        self.ui.rdoMeasure.setEnabled(False)
        self.ui.rdoPresence.setChecked(False)
        self.ui.rdoPresence.setEnabled(False)
        self.ui.rdoCalculate.setChecked(False)
        self.ui.rdoCalculate.setEnabled(False)
        # disable multi select options
        self.setPresence()
        # enable selection of calculation field
        self.ui.lblCalcField.setEnabled(True)
        self.ui.lblCalcField.setText('Select field to create new fields for each unique value')
        self.ui.cbxCalcField.setEnabled(True)
        # disable re-use of an existing field
        self.ui.lblSelectField.setEnabled(False)
        self.ui.cbxSelectField.setEnabled(False)
        self.ui.lblNewField.setText('Enter field prefix (up to 6 characters)')
        self.ui.leNewField.setMaxLength(6)
        
    def setCalculate(self):
        self.ui.lblCalcField.setEnabled(True)
        self.ui.cbxCalcField.setEnabled(True)
        if not self.ui.lblMultipleAction.isEnabled():
            self.ui.lblMultipleAction.setEnabled(True)
            self.ui.rdoSum.setEnabled(True)
            self.ui.rdoMean.setEnabled(True)
            self.ui.rdoMax.setEnabled(True)
            self.ui.rdoMin.setEnabled(True)
            self.ui.rdoCount.setEnabled(True)
    
    def setPresence(self):
        self.ui.lblCalcField.setDisabled(True)
        self.ui.cbxCalcField.setDisabled(True)
        if self.ui.lblMultipleAction.isEnabled():
            self.ui.lblMultipleAction.setDisabled(True)
            self.ui.rdoSum.setDisabled(True)
            self.ui.rdoMean.setDisabled(True)
            self.ui.rdoMax.setDisabled(True)
            self.ui.rdoMin.setDisabled(True)
            self.ui.rdoCount.setDisabled(True)

    def updateMeasureLayer(self, layers):
        self.ui.cbxMeasureLayer.clear()
        self.ui.cbxMeasureLayer.addItems(layers)
        self.setCalcFields()

    def setCalcFields(self):
        import ftools_utils
        self.ui.cbxCalcField.clear()
        calcLayer = self.ui.cbxMeasureLayer.currentText()
        selectedLayer = ftools_utils.getVectorLayerByName(unicode(calcLayer))
        if selectedLayer != None:
            fields = ftools_utils.getFieldList(selectedLayer)
            for i in fields:
                if fields[i].type() == QtCore.QVariant.Int or \
                fields[i].type() == QtCore.QVariant.Double:
                    self.ui.cbxCalcField.addItem(unicode(fields[i].name()))
    
    def setMeasurePoints(self):
        import ftools_utils
        layers = ftools_utils.getLayerNames([QGis.Point])
        self.ui.chkCreateTemp.setEnabled(False)
        self.ui.chkCreateTemp.setChecked(False)
        self.updateMeasureLayer(layers)
        
    def setMeasureLines(self):
        import ftools_utils
        layers = ftools_utils.getLayerNames([QGis.Line])
        self.ui.chkCreateTemp.setEnabled(True)
        self.updateMeasureLayer(layers)
        
    def setMeasurePolygons(self):
        import ftools_utils
        layers = ftools_utils.getLayerNames([QGis.Polygon])
        self.ui.chkCreateTemp.setEnabled(True)
        self.updateMeasureLayer(layers)
    
    def setMeasureFields(self, planningGrid):
        import ftools_utils
        self.ui.cbxSelectField.clear()
        selectedLayer = ftools_utils.getVectorLayerByName(unicode(planningGrid))
        if selectedLayer != None:
            self.ui.cbxSelectField.addItem(unicode("--Create New--"))
            fields = ftools_utils.getFieldList(selectedLayer)
            for i in fields:
                if fields[i].type() == QtCore.QVariant.Int or \
                fields[i].type() == QtCore.QVariant.Double:
                    self.ui.cbxSelectField.addItem(unicode(fields[i].name()))
    
    def setOutputField(self):
        if self.ui.cbxSelectField.currentIndex() == 0:
            self.ui.lblNewField.setEnabled(True)
            self.ui.leNewField.setEnabled(True)
        else:
            self.ui.lblNewField.setDisabled(True)
            self.ui.leNewField.setDisabled(True)
          
    def accept(self):            
        import ftools_utils
        self.ui.buttonOk.setEnabled(False)
        puFile = self.ui.cbxPlanningGrid.currentText()
        mFile = self.ui.cbxMeasureLayer.currentText()
        if self.ui.chkSelectedOnly.isChecked():
            useSelected = True
        else:
            useSelected = False
        proceed = True
        if self.ui.cbxSelectField.currentIndex() == 0:
            if self.ui.leNewField.text() != "":
                fName = self.ui.leNewField.text()
                proceed = True
            else:
                proceed = False
        else:
            fName = self.ui.cbxSelectField.currentText()
        if proceed:
            self.ui.progressBar.setRange( 0, 100 )
            self.do_calculation(puFile, mFile, fName, useSelected)

        # reset progress bar
        self.ui.progressBar.setValue(0)
        # reset system
        self.ui.buttonOk.setEnabled(True)
            
    #
    # Functions to create temporary file simplified from qm_mkgrid
    #
    
    def CreateSquare(self, x, y, sideLen):
    
        pt1 = QgsPoint(x, y)
        pt2 = QgsPoint(x + sideLen, y)
        pt3 = QgsPoint(x + sideLen, y - sideLen)
        pt4 = QgsPoint(x, y - sideLen)
        pt5 = QgsPoint(x, y)
        square = [[pt1, pt2, pt3, pt4, pt5]]
        return(square)
        
    def writeFile(self, bound, sideLen, shapefileName, encoding):
        
        import ftools_utils
        crs = self.iface.mapCanvas().mapRenderer().destinationSrs()
        if not crs.isValid():
            crs = None
        fields = {0:QgsField("temp_id", QtCore.QVariant.Int)}
        check = QtCore.QFile(shapefileName)
        if check.exists():
            if not QgsVectorFileWriter.deleteShapeFile(shapefileName):
                return
        writer = QgsVectorFileWriter(shapefileName, encoding, fields, QGis.WKBPolygon, crs)
        outFeat = QgsFeature()
        outGeom = QgsGeometry()
        idVar = 0
        # place squares from top left corner
        y = bound.yMaximum()
        rowCount = int((bound.yMaximum()-bound.yMinimum())/sideLen)
        row = 1
        while y >= bound.yMinimum():
            x = bound.xMinimum()
            while x <= bound.xMaximum():
                polygon = self.CreateSquare(x, y, sideLen)
                outFeat.setGeometry(outGeom.fromPolygon(polygon))
                outFeat.addAttribute(0, QtCore.QVariant(idVar))
                writer.addFeature(outFeat)
                idVar = idVar + 1
                x = x + sideLen
            y = y - sideLen
        del writer

    def intersectLayers(self, gridLyr, srcLyr, outName, calcindex, selectedIds, useSelected):

        import ftools_utils
        crs = self.iface.mapCanvas().mapRenderer().destinationSrs()
        if not crs.isValid():
            crs = None
        if calcindex < 0:
            fields = {0:QgsField("temp_id", QtCore.QVariant.Int)}
        else:
            srcFields = ftools_utils.getFieldList(srcLyr)
            fields = {0:QgsField("temp_id", QtCore.QVariant.Int),
                      1:QgsField(srcFields[calcindex].name(),srcFields[calcindex].type() )}
        check = QtCore.QFile(outName)
        if check.exists():
            if not QgsVectorFileWriter.deleteShapeFile(outName):
                return
        if self.ui.rdoLines.isChecked():
            writer = QgsVectorFileWriter(outName, gridLyr.dataProvider().encoding(), fields, QGis.WKBLine, crs)
        else:
            writer = QgsVectorFileWriter(outName, gridLyr.dataProvider().encoding(), fields, QGis.WKBPolygon, crs)
        outFeat = QgsFeature()
        outGeom = QgsGeometry()
        idVar = 0
        # index source layer
        srcFeat = QgsFeature()
        srcGeom = QgsGeometry()
        try:
            index = QgsSpatialIndex()
            srcLyr.select([], QgsRectangle(), True, False)
            while srcLyr.nextFeature(srcFeat):
                index.insertFeature(srcFeat)
        except:
            return(False,'Could not build index, possible topology errors')
        # set inputs
        gridFeat = QgsFeature()
        gridGeom = QgsGeometry()
        gridProvider = gridLyr.dataProvider()
        fCount = gridProvider.featureCount()
        x = 0 # id for intersected features
        y = 0 # number of pu's processed
        try:
            while gridProvider.nextFeature( gridFeat ):
                self.ui.progressBar.setValue((y*100)/fCount/2)
                self.ui.progressBar.repaint()
                # get geometry from current feature in grid layer
                gridGeom = QgsGeometry( gridFeat.geometry() )
                # get list of features in source layer that might intersect
                features = index.intersects( gridGeom.boundingBox() )
                # is there anything to intersect?
                if len(features) > 0:
                    for fid in features:
                        if useSelected: # should we only use the selected ones?
                            if not id in selectedIds:
                                continue # skip this one...
                        # get an input layer feature
                        srcLyr.featureAtId(fid, srcFeat, True)
                        # get the input layer feature's geometry
                        try:
                            srcGeom = QgsGeometry( srcFeat.geometry() )
                            outGeom = QgsGeometry( gridGeom.intersection(srcGeom) )
                            outFeat.setGeometry( outGeom )
                            outFeat.addAttribute(0,x)
                            if calcindex >= 0:
                                attMap = srcFeat.attributeMap()
                                outVal = attMap[calcindex]
                                outFeat.addAttribute(1,outVal)
                            x = x + 1
                            writer.addFeature(outFeat)
                        except:
                            return(False,'Measure layer has invalid topology. Correct before continuing.')
                y = y + 1
        except Exception, err:
            message = "An error occured while creating temporary file:\n%s" % str(err)
            return (False, message)
        del writer

    #
    # Derived from fTools point in polygon function
    #
        
    def do_calculation(self, planningUnitFile, measureFile, newFieldName, useSelected):
        
        import ftools_utils
        # define layers
        puLayer = ftools_utils.getVectorLayerByName(planningUnitFile)
        measureLayer = ftools_utils.getVectorLayerByName(measureFile)
        # get selected IDs
        if useSelected:
            selectedIds = measureLayer.selectedFeaturesIds()
        else:
            selectedIds = []
        fields = ftools_utils.getFieldList(measureLayer)
        # sel calculation field name
        cfn = self.ui.cbxCalcField.currentText()
        # determine actions
        # default for value is measure
        calcindex = -1
        multiCalcIndex = -1
        # default for multiple intersection is sum
        multiAction = 'presence'
        if not self.ui.rdoMeasure.isChecked():
            for i in fields:
                if fields[i].name() == cfn:
                    calcindex = i
        if self.ui.rdoMean.isEnabled():
            if self.ui.rdoMean.isChecked():
                multiAction = 'mean'
            elif self.ui.rdoMax.isChecked():
                multiAction = 'max'
            elif self.ui.rdoMin.isChecked():
                multiAction = 'min'
            elif self.ui.rdoCount.isChecked():
                multiAction = 'count'
            else:
                multiAction = 'sum'
        if self.ui.rdoCalcMulti.isChecked():
            multiAction = 'sumbyid'
            for i in fields:
                if fields[i].name() == cfn:
                    multiCalcIndex = i
        if measureLayer.geometryType() == QGis.Point:
            if newFieldName == "":
                fieldName = QtCore.QString("pnt_cnt")
            else:
                fieldName = newFieldName
        elif measureLayer.geometryType() == QGis.Line:
            if newFieldName == "":
                fieldName = QtCore.QString("line_len")
            else:
                fieldName = newFieldName
        elif measureLayer.geometryType() == QGis.Polygon:
            if newFieldName == "":
                fieldName = QtCore.QString("poly_area")
            else:
                fieldName = newFieldName
        else:
            message = "Invalid input layer type: Should be point,line, or polygon layer"
            return (False, message)
        # determine if we should create a temporary layer
        if self.ui.chkCreateTemp.isChecked():
            # only do this for lines and polygons because waste of time for points
            # get extents of current measure layer
            boundBox = measureLayer.extent()
            # create at least 25 x 25 box
            # get side length
            sideLen = min( ( (boundBox.xMaximum() - boundBox.xMinimum()) / 25) ,
                ( (boundBox.yMaximum() - boundBox.yMinimum()) / 25) )
            self.writeFile(boundBox, sideLen, 'tempgrid.shp', measureLayer.dataProvider().encoding())
            # add to view
            gridLayer = qgis.utils.iface.addVectorLayer('tempgrid.shp','','ogr')
            # do intersection
            if self.ui.rdoCalcMulti.isChecked() == True:
                self.intersectLayers(gridLayer, measureLayer, 'tempints.shp', multiCalcIndex, selectedIds, useSelected)
            else:
                self.intersectLayers(gridLayer, measureLayer, 'tempints.shp', calcindex, selectedIds, useSelected)
            # remove grid layer from view
            QgsMapLayerRegistry.instance().removeMapLayers( [gridLayer.id()] )
            # remove grid from file system
            os.remove('tempgrid.shp')
            os.remove('tempgrid.dbf')
            os.remove('tempgrid.shx')
            os.remove('tempgrid.prj')
            os.remove('tempgrid.qpj')
            # load intersection
            tempLayer = qgis.utils.iface.addVectorLayer('tempints.shp','','ogr')
            # do calculation
            success = False
            message = 'Unexpected Failure'
            if calcindex >= 0:
                success, message = self.calcFeaturesInPU(puLayer, tempLayer, fieldName, \
                    multiAction, selectedIds, 1, cfn, useSelected)
            else:
                success, message = self.calcFeaturesInPU(puLayer, tempLayer, fieldName, \
                    multiAction, selectedIds, calcindex, cfn, useSelected)
            self.ui.progressBar.setValue(50)
            self.ui.progressBar.repaint()
            # remove temporary layers from view
            QgsMapLayerRegistry.instance().removeMapLayers( [tempLayer.id()] )
            # remove temporary layers from file system
            ## temp intersect
            os.remove('tempints.shp')
            os.remove('tempints.dbf')
            os.remove('tempints.shx')
            os.remove('tempints.prj')
            os.remove('tempints.qpj')
        else:
            success, message = self.calcFeaturesInPU(puLayer, measureLayer, fieldName, \
                multiAction, selectedIds, calcindex, cfn, useSelected)
        if success:
            self.ui.progressBar.setValue(100)
            self.ui.progressBar.repaint()
            QtGui.QMessageBox.information(self, self.tr("Planning Unit Calculation"), 
                        self.tr(message))
        else:
            QtGui.QMessageBox.warning(self, self.tr("Planning Unit Calculation"), 
                        self.tr(message))


    def getUniqueValues(self, measureLayer, fieldname):

        # create data provider
        prov = measureLayer.dataProvider()
        # select only the attribute we want
        fidx = prov.fieldNameIndex(fieldname)
        # don't return geometry - not needed
        prov.select([fidx], QgsRectangle(), False)
        # add values for all items
        idList = []
        feat = QgsFeature()
        while prov.nextFeature(feat):
            attrs = feat.attributeMap()
            for (k,attr) in attrs.iteritems():
                idList.append(str(attr.toString()))
        # create unique list using set creation        
        uniqueIds = list(set(idList))
        # convert to integers
        intIds = []
        for i in uniqueIds:
            intIds.append(int(i))
        # sort integer list
        intIds.sort()
        # return values
        return(intIds)
        
    #
    # TSW 2011-11-15
    # The following function slightly modified from
    # a script posted by Carson Farmer. New field name
    # selection has been added. Als returns of error messages
    # has been added to allow for use in GUI.
    #

    #-----------------------------------------------------------
    #
    # This file is part of fTools
    # Copyright (C) 2011  Carson Farmer
    # EMAIL: carson.farmer (at) gmail.com
    # WEB  : http://www.ftools.ca/fTools.html
    #
    # A collection of data management and analysis tools for vector data
    #
    #-----------------------------------------------------------
    #
    # licensed under the terms of GNU GPL 2
    #
    # This program is free software; you can redistribute it and/or modify
    # it under the terms of the GNU General Public License as published by
    # the Free Software Foundation; either version 2 of the License, or
    # (at your option) any later version.
    #
    # This program is distributed in the hope that it will be useful,
    # but WITHOUT ANY WARRANTY; without even the implied warranty of
    # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    # GNU General Public License for more details.
    #
    # You should have received a copy of the GNU General Public License along
    # with this program; if not, write to the Free Software Foundation, Inc.,
    # 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
    #
    #---------------------------------------------------------------------

    def calcFeaturesInPU(self, puLayer, inputLayer, fieldName, multiAction, selectedIds, calcindex, calcFieldName, onlySelected=False):

        # import required library
        import ftools_utils
        # set defaults
        message = "Calculation Complete!"
        feature = QgsFeature()
        # lock layer
        puLayer.blockSignals(True)
        wasEditing = puLayer.isEditable() # if it is editable, then we must be in edit mode
        if not wasEditing:
            puLayer.startEditing()
            if not puLayer.isEditable():
                message = "Unable to edit input pu layer: Please choose a layer with edit capabilities."
                return (False, message)
        if not inputLayer.crs() == puLayer.crs():
            message = "Input layers have non-matching CRS. Can not continue."
            return (False, message)
        # build field map & determine if new fields are needed
        fieldMap = puLayer.pendingFields()
        fieldMap = dict([(field.name(), index) for index, field in fieldMap.iteritems()])
        attributeId = -1
        if fieldName.toUpper() in fieldMap: # update existing uppercase field
            attributeId = fieldMap[fieldName.toUpper()]
        elif fieldName in fieldMap: # update existing lowercase field
            attributeId = fieldMap[fieldName]
        elif self.ui.rdoCalcMulti.isChecked() == True:
            # this allows users to recalculate or update without having to
            # delete fields
            uniqueIds = self.getUniqueValues(inputLayer, calcFieldName)
            newIds = []
            for uID in uniqueIds:
                nfn = fieldName+str(uID)
                if nfn.toUpper() in fieldMap:
                    attributeId = fieldMap[nfn.toUpper()]
                elif nfn in fieldMap:
                    attributeId = fieldMap[nfn]
                else:
                    newIds.append(uID)
            if len(newIds) > 0:
                attributeId = -1
        # create new fields
        if attributeId == -1: # create new field
            if not puLayer.dataProvider().capabilities() > 7: # can't add attributes
                message = "Data provider does not support adding attributes: " + \
                    "Cannot add required field."
                if not wasEditing:
                    puLayer.rollBack()
                return (False, message)
            # determine if one field is to be created or multiple fields
            if self.ui.rdoCalcMulti.isChecked() == True:
                # create multiple fields
                for uID in newIds:
                    nfn = fieldName+str(uID)
                    newField = QgsField(nfn, QtCore.QVariant.Double, "Real", 19, 10)
                    puLayer.beginEditCommand("Attribute added")
                    if not puLayer.addAttribute(newField):
                        message = "Could not add new fields to the PU layer"
                        puLayer.destroyEditCommand()
                        if not wasEditing:
                            puLayer.rollBack()
                        return(False, message)
                    puLayer.endEditCommand()
                # get index of the new field
                fieldMap = puLayer.pendingFields()
                fieldMap = dict([(field.name(), index) for index, field in fieldMap.iteritems()])
                if not nfn in fieldMap:
                    message =  "Could not find the newly created field."
                    if not wasEditing:
                        puLayer.rollBack()
                    return (False, message)
                attributeId = fieldMap[nfn]
            else:
                # create a single field
                newField = QgsField(fieldName, QtCore.QVariant.Double, "Real", 19, 10)
                puLayer.beginEditCommand("Attribute added")
                if not puLayer.addAttribute(newField):
                    message = "Could not add the new field to the PU layer."
                    puLayer.destroyEditCommand()
                    if not wasEditing:
                        puLayer.rollBack()
                    return (False, message)
                puLayer.endEditCommand()
                # get index of the new field
                fieldMap = puLayer.pendingFields()
                fieldMap = dict([(field.name(), index) for index, field in fieldMap.iteritems()])
                if not fieldName in fieldMap:
                    message =  "Could not find the newly created field."
                    if not wasEditing:
                        puLayer.rollBack()
                    return (False, message)
                attributeId = fieldMap[fieldName]
        if attributeId == -1:
            message= "Error writing to the new attribute."
            if not wasEditing:
                puLayer.rollBack()
            return (False, message)
        #        
        # commit changes to table structure
        puLayer.blockSignals(False)
        puLayer.setModified(True, False)
        if not wasEditing:
            puLayer.commitChanges()
        #
        # put back into edit state
        puLayer.blockSignals(True)
        wasEditing = puLayer.isEditable() # if it is editable, then we must be in edit mode
        if not wasEditing:
            puLayer.startEditing()
            if not puLayer.isEditable():
                message = "Unable to edit input pu layer: Please choose a layer with edit capabilities."
                return (False, message)
        #
        # create spatial index input layer for faster querying later
        try:
            index = QgsSpatialIndex()
            inputLayer.select([], QgsRectangle(), True, False)
            while inputLayer.nextFeature(feature):
                index.insertFeature(feature)
        except:
            return(False,'Could not build index, possible topology errors')
        # do the calculation
        puLayer.select(puLayer.pendingAllAttributesList(), QgsRectangle(), True, False)
        puLayer.beginEditCommand("Count field updated")
        inputFeature = QgsFeature()
        measure = QgsDistanceArea()
        geom = QgsGeometry()
        fCount = puLayer.dataProvider().featureCount()
        x = 0
        try:
            #QtGui.QMessageBox.information(self, self.tr("Debug"), self.tr('calcindex = %d' % calcindex))
            while puLayer.nextFeature(feature):
                if self.ui.chkCreateTemp.isChecked():
                    self.ui.progressBar.setValue(((x*100)/fCount/2) + 50)
                else:
                    self.ui.progressBar.setValue((x*100)/fCount)
                self.ui.progressBar.repaint()
                # get geometry from current feature in polygon layer
                geom = QgsGeometry(feature.geometry())
                # get list of features in input layer that might intersect
                features = index.intersects(geom.boundingBox())
                # set initial values
                value = 0.0
                valMax = None
                valMin = None
                valCount = 0.0
                featureVal = None
                # is there anything to calculate?
                if len(features) > 0:
                    if self.ui.rdoCalcMulti.isChecked() == True:
                        idIdx = inputLayer.dataProvider().fieldNameIndex(calcFieldName)
                        idValues = {}
                        for id in features:
                            if onlySelected: # should we only use the selected ones?
                                if not id in selectedIds:
                                    continue # skip this one...
                            # get an input layer feature attributes & geometry
                            inputLayer.featureAtId(id, inputFeature, True)
                            attMap = inputFeature.attributeMap()
                            inputGeom = QgsGeometry(inputFeature.geometry())
                            # determine geographic type and act accordingly
                            key = str(attMap[idIdx].toString())
                            if inputLayer.geometryType() == QGis.Point:
                                if geom.contains(inputGeom.asPoint()):
                                    if key in idValues:
                                        idValues[key] += 1
                                    else:
                                        idValues[key] = 1
                            else:
                                try:
                                    inputGeom = QgsGeometry(geom.intersection(inputGeom))
                                except:
                                        return(False,'Measure layer has invalid topology. Correct before continuing.')
                                if not inputGeom.isGeosValid() or inputGeom.isGeosEmpty():
                                    continue
                                value = measure.measure(inputGeom)
                                if key in idValues:
                                    idValues[key] += value
                                else:
                                    idValues[key] = value
                        # set value to PU
                        for fid in uniqueIds:
                            nfn = fieldName+str(fid)
                            idx = fieldMap[nfn]
                            if str(fid) in idValues:
                                value = idValues[str(fid)]
                            else:
                                value = 0.0
                            puLayer.changeAttributeValue(feature.id(), int(idx), value, False)
                    else:
                        if calcindex < 0:
                            # direct measure or presence
                            for id in features:
                                if onlySelected: # should we only use the selected ones?
                                    if not id in selectedIds:
                                        continue # skip this one...
                                # get an input layer feature
                                inputLayer.featureAtId(id, inputFeature, True)
                                # get the input layer feature's geometry
                                inputGeom = QgsGeometry(inputFeature.geometry())
                                if inputLayer.geometryType() == QGis.Point:
                                    if geom.contains(inputGeom.asPoint()):
                                        value += 1.0
                                        valCount += 1.0
                                        valMax = 1.0
                                        valMin = 1.0
                                else:
                                    try:
                                        inputGeom = QgsGeometry(geom.intersection(inputGeom))
                                    except:
                                            return(False,'Measure layer has invalid topology. Correct before continuing.')
                                    if not inputGeom.isGeosValid() or inputGeom.isGeosEmpty():
                                        continue
                                    value += measure.measure(inputGeom)
                                    valCount += 1.0
                                    if valMax is None or measure.measure(inputGeom) > valMax:
                                        valMax = measure.measure(inputGeom)
                                    if valMin is None or measure.measure(inputGeom) < valMin:
                                        valMin = measure.measure(inputGeom)
                        else:
                            # measure * field
                            for id in features:
                                if onlySelected: # should we only use the selected ones?
                                    if not id in selectedIds:
                                        continue # skip this one...
                                # get an input layer feature
                                inputLayer.featureAtId(id, inputFeature, True)
                                # get the input layer feature's geometry
                                inputGeom = QgsGeometry(inputFeature.geometry())
                                if inputLayer.geometryType() == QGis.Point:
                                    if geom.contains(inputGeom.asPoint()):
                                        attMap = inputFeature.attributeMap()
                                        value += float(attMap[calcindex].toString())
                                        valCount += 1.0
                                        if valMax is None or float(attMap[calcindex].toString()) > valMax:
                                            valMax = float(attMap[calcindex].toString())
                                        if valMin is None or float(attMap[calcindex].toString()) < valMin:
                                            valMin = float(attMap[calcindex].toString())
                                else:
                                    try:
                                        inputGeom = QgsGeometry(geom.intersection(inputGeom))
                                    except:
                                            return(False,'Measure layer has invalid topology. Correct before continuing.')
                                    if not inputGeom.isGeosValid() or inputGeom.isGeosEmpty():
                                        continue
                                    attMap = inputFeature.attributeMap()
                                    value += float(attMap[calcindex].toString()) * measure.measure(inputGeom)
                                    valCount += 1.0
                                    if valMax is None or float(attMap[calcindex].toString()) * measure.measure(inputGeom) > valMax:
                                        valMax = float(attMap[calcindex].toString()) * measure.measure(inputGeom)
                                    if valMin is None or float(attMap[calcindex].toString()) * measure.measure(inputGeom) < valMin:
                                        valMin = float(attMap[calcindex].toString()) * measure.measure(inputGeom)
                        if multiAction == 'sum':
                            featureVal = value
                        elif multiAction == 'mean':
                            if value > 0:
                                featureVal = value / valCount
                            else:
                                featureVal = value 
                        elif multiAction == 'max':
                            featureVal = valMax
                        elif multiAction == 'min':
                            featureVal = valMin
                        elif multiAction == 'count':
                            featureVal = valCount
                        elif multiAction == 'presence':
                            if valCount > 0.0:
                                featureVal = 1.0
                            else:
                                featureVal = 0.0
                        # set value to PU 
                        if featureVal is None:
                            featureVal = 0.0
                        puLayer.changeAttributeValue(feature.id(), attributeId, featureVal, False)
                else:
                    if self.ui.rdoCalcMulti.isChecked() == True:
                        for fid in uniqueIds:
                            nfn = fieldName+str(fid)
                            idx = fieldMap[nfn]
                            puLayer.changeAttributeValue(feature.id(), int(idx), 0.0, False)
                    else:
                        puLayer.changeAttributeValue(feature.id(), attributeId, 0.0, False)
                x = x + 1
            puLayer.endEditCommand()
        except Exception, err:
            message = "An error occured while adding measure:\n%s" % str(err)
            if not wasEditing:
                puLayer.rollBack()
                return (False, message)

        # stop blocking layerModified signals and make sure that one layerModified signal is emitted
        puLayer.blockSignals(False)
        puLayer.setModified(True, False)

        if not wasEditing:
            puLayer.commitChanges()
        return (True, message)
